class CFG(object):

    def __init__(self, bytecode=None, instructions=None, analysis=True):
        """ TODO """
        raise NotImplementedError

    def show(self):
        """ TODO """
        raise NotImplementedError
